#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "config_xor.h"

#ifdef SH_STEALTH
char * globber(char * string);
#define _(string) globber(string) 
#define N_(string) string
#else
#define _(string)  string 
#define N_(string) string
#endif

#ifdef SH_STEALTH
#ifndef SH_MAX_GLOBS
#define SH_MAX_GLOBS 32
#endif
char * globber(char * str)
{
  register int i, j;
  static int  count = -1;
  static char glob[SH_MAX_GLOBS][128];

  ++count; if (count > (SH_MAX_GLOBS-1) ) count = 0;
  j = strlen(str);
  if (j > 127) j = 127;

  for (i = 0; i < j; ++i)
    {
      if (str[i] != '\n' && str[i] != '\t') 
	glob[count][i] = str[i] ^ XOR_CODE;
      else
	glob[count][i] = str[i];
    }
  glob[count][j] = '\0';
  return glob[count];
}
#endif

/* This is a very inefficient algorithm, but there is really no
 * need for anything more elaborated here. Can handle NULL's in haystack
 * (not in needle), which strstr() apparently cannot.
 */
char * my_strstr (char * haystack, char * needle, int haystack_size)
{
  register int      i = 0, j = 0;
  register int      siz;
  register char * ptr = haystack;
  register int      len;

  siz = strlen(needle);
  len = haystack_size - siz;

  while (j < len)
    {
      i = 0;
      while (i < siz)
	{
	  if (needle[i] != ptr[i]) break;
	  if (i == (siz-1)) 
	      return ptr;
	  ++i;
	}
      ++ptr; ++j;
    }
  return NULL;
}

/* fread()  does not return the number of chars read, thus we need to
 * read only a small number of bytes, in order not to expand the binary
 * too much with the last fwrite(). Too lazy to fix this now. 
 */
#define GRAB_SIZE 32

int readhexchar ( char c )
{
  if      ( c >= '0' && c <= '9' )
    return c - '0';
  else if ( c >= 'a' && c <= 'f' )
    return c - 'a' + 10;
  else if ( c >= 'A' && c <= 'F' )
    return c - 'A' + 10;
  else return -1;
}

int main (int argc, char * argv[])
{
  /* the default password
   */
  char TcpFlag[9] = { 0xF7,0xC3,0x12,0xAA,0xAA,0x12,0xC3,0xF7 }; 
  
  char * found_it;
  int    i;
  int    suc = 0;
  int    count = 0;
  char * newn;
  FILE * oldf;
  FILE * newf;

  char   in[9];
  int    j, k;
  char * str;

  char * buf = (char *) malloc(GRAB_SIZE);
  size_t dat;
  char * newpwd = (char *) malloc(5 * 8 + 2); 
  char * oldpwd = (char *) malloc(5 * 8 + 2); 


  if (argc < 4) 
    {
      fprintf (stderr, _("\nUsage: samhain_setpwd <filename> <suffix> "\
	       "<new_password>\n\n"));
      fprintf (stderr, _("   This program is a utility that will:\n"));
      fprintf (stderr, _("    - search in the binary executable <filename> "\
	       "for samhain's\n"));
      fprintf (stderr, _("      compiled-in default password,\n"));
      fprintf (stderr, _("    - change it to <new_password>,\n"));
      fprintf (stderr, _("    - and output the modified binary to "\
	       "<filename>.<suffix>\n\n"));
      fprintf (stderr, _("   To allow for non-printable chars, "\
			 "<new_password> must be\n")); 
      fprintf (stderr, _("   a 16-digit hexadecimal "\
	       "number (only 0-9,A-F allowed in input),\n"));
      fprintf (stderr, _("   thus corresponding"\
			 "   to an 8-byte password.\n\n"));
      fprintf (stderr, _("   Example: 'samhain_setpwd samhain new "\
	       "4142434445464748'\n"));
      fprintf (stderr, _("   takes the file 'samhain', sets the password to "\
	       "'ABCDEFGH'\n")); 
      fprintf (stderr, _("   ('A' = 41 hex, 'B' = 42 hex, ...) "\
	       "and outputs the result\n"));
      fprintf (stderr, _("   to 'samhain.new'.\n"));
      return -1;
    }

  if (strlen(argv[3]) != 16)
    {
      fprintf (stderr, _("<new_password> %s has not exactly 16 chars\n"),
	       argv[0]);
      return -1;
    }
  str = &argv[3][0];
  i = 0;
  while (i < 16)
    {
      k = i/2; j = 0; 
      if (2*k == i) in[k] = 0;
      while (j < 16)
	{
	  if (-1 != readhexchar(str[i])) 
	    {
	      in[k] += readhexchar(str[i]) * (i == 2*k ? 16 : 1);
	      break;
	    }
	  ++j;
	  if (j == 16) 
	    {
	      fprintf(stderr, _("Invalid char %c\n"), str[i]);
	      _exit(EXIT_FAILURE);
	    }
	}
      ++i;
    }
  in[8] = '\0';

  /* fprintf(stderr, "%s\n", in); */
    
  oldf = fopen(argv[1], "r");
  newn = (char *) malloc (strlen(argv[1])+strlen(argv[2])+2);
  strcpy(newn, argv[1]);
  strcat(newn, ".");
  strcat(newn, argv[2]);
  newf = fopen(newn, "w");

  /* fprintf(stderr, _("Output to %s\n"), newn); */

  while (1)
    {
      dat = fread (buf, GRAB_SIZE, 1, oldf); ++count;
      if (dat == 0) break;
      if ( (found_it = my_strstr(buf, TcpFlag, GRAB_SIZE)) != NULL)
	{
	  suc = 1;
	  fprintf (stderr, _(".. old password found\n"));
	  for (i = 0; i < 8; ++i)
	    {
              /* not a nicely formatted output. who cares.
               */
	      if (!isgraph((int) *found_it))
		{
		  sprintf(&oldpwd[i*4], _("%c%03o"), 
			  '\\',  (unsigned char) *found_it);
		}
	      else
		{
		  sprintf(&oldpwd[i*4], _("%c   "), *found_it);
		}
	      if (!isgraph((int) in[i]))
		sprintf(&newpwd[i*4], _("%c%03o"), 
			'\\',  (unsigned char) in[i]);
	      else
		sprintf(&newpwd[i*4], _("%c   "), in[i]);
	      *found_it = in[i];
	      ++found_it;
	    }
	  fprintf (stderr, _("Replaced:  %s  by:  %s\n"), oldpwd, newpwd);
	}
      fwrite (buf, GRAB_SIZE, 1, newf);
    }
  fwrite (buf, GRAB_SIZE, 1, newf);
  fprintf (stderr, _(".. finished\n"));
  fclose (newf);
  fclose (oldf);

  if (suc == 0)
    {
      fprintf (stderr, _("Not found in first pass.\n"));
      fprintf (stderr, _("Second pass ..\n"));
      count = 0;
      oldf  = fopen(argv[1], "r");
      newf  = fopen(newn, "w");
      dat   = fread (buf, 10, 1, oldf);
      fwrite (buf, 10, 1, newf);
      while (1)
	{
	  dat = fread (buf, GRAB_SIZE, 1, oldf); ++count;
	  if (dat == 0) break;
	  if ( (found_it = my_strstr(buf, TcpFlag, GRAB_SIZE)) != NULL)
	    {
	      suc = 1;
	      fprintf (stderr, _(".. old password found\n"));
	      for (i = 0; i < 8; ++i)
		{
		  if (!isgraph((int) *found_it))
		    sprintf(&oldpwd[i*4], "\\%03o", (unsigned char)*found_it);
		  else
		    sprintf(&oldpwd[i*4], "%c   ", *found_it);
		  if (!isgraph((int) in[i]))
		    sprintf(&newpwd[i*4], "\\%03o", (unsigned char)in[i]);
		  else
		    sprintf(&newpwd[i*4], "%c   ", in[i]);
		  *found_it = in[i];
		  ++found_it;
		}
	      fprintf (stderr, _("Replaced %s by %s\n"), oldpwd, newpwd);
	    }
	  fwrite (buf, GRAB_SIZE, 1, newf);
	}
      fwrite (buf, GRAB_SIZE, 1, newf);
      fprintf (stderr, _(".. finished second pass\n"));
      fclose (newf);
      fclose (oldf);
    }

  return 0;
}
