/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */


#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "config_xor.h"

#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif

#ifdef TM_IN_SYS_TIME
#include <sys/time.h>
#else
#include <time.h>
#endif

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#include <sys/types.h>

#ifdef  HAVE_UNISTD_H
#include <errno.h>
#include <signal.h>
#include <setjmp.h>
#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <unistd.h>
#endif

#ifndef FD_SET
#define NFDBITS         32
#define FD_SETSIZE      32
#define FD_SET(n, p)    ((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define FD_CLR(n, p)    ((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define FD_ISSET(n, p)  ((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#define FD_ZERO(p)      memset((char *)(p), '\0', sizeof(*(p)))
#endif 

#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
#include <sys/mman.h>
#endif


#include <netdb.h> 
#include <sys/types.h> 
#include <netinet/in.h> 
#include <sys/socket.h> 
#include <arpa/inet.h>

#include "samhain.h"
#include "sh_tiger.h"
#include "sh_utils.h"
#include "sh_forward.h"
#include "sh_srp.h"
#include "sh_fifo.h"


/* the port client will be connecting to 
 */
#ifndef SH_DEFAULT_PORT
#define SH_DEFAULT_PORT 49777    
#endif

#ifndef SH_SELECT_REPEAT
#define SH_SELECT_REPEAT 60
#endif

#ifndef SH_HEADER_SIZE
#define SH_HEADER_SIZE 7
#endif

#ifndef SH_CHALLENGE_SIZE
#define SH_CHALLENGE_SIZE 9
#endif

/* protocols
 */
#define SH_PROTO_SRP 0x01
#define SH_PROTO_MSG 0x02

#ifndef SH_STANDALONE

extern int h_errno;

char * sh_forward_errmessage (int tellme)
{

#ifdef HOST_NOT_FOUND
    if (tellme == HOST_NOT_FOUND)  
      return _("The specified host is unknown.");
#endif
#ifdef NO_ADDRESS
    if (tellme == NO_ADDRESS)  
      return _("The requested name is valid but does not have an IP address.");
#endif
#ifdef NO_RECOVERY
    if (tellme == NO_RECOVERY)  
      return _("A non-recoverable name server error occurred.");
#endif
#ifdef TRY_AGAIN
    if (tellme == TRY_AGAIN)  
      return _("A temporary error occurred on an authoritative name server.The specified host is unknown.");
#endif
    return _("Unknown error.");
}


static int StripDomain = GOOD;

int sh_forward_set_strip (char * dummy)
{
  StripDomain = BAD;
  return GOOD;
}

#include <ctype.h>

const char *sh_strip_domain (char *name)
{
  char *          first;
  static char     name_2[SH_MINIBUF+1];
  register int    i = 0;
  register char * testit;
  int             testflag;

  if (StripDomain == BAD || (first  = strchr(name, '.')) == NULL) 
    return name;
  else
    {

      /* check whether it is in dotted number format
       */
      testit = &name[0]; testflag = GOOD;
      while (*testit != '\0') 
	{
	  if (!isdigit((int) *testit) && *testit != '.')
	    {
	      testflag = BAD;
	      break;
	    } 
	  ++testit;
	}

      if (testflag == BAD)
	{
	  first = name;
	  while (i < SH_MINIBUF && *first != '.' && *first != '\0')
	    {
	      name_2[i] = *first;
	      ++first; ++i;
	    }
	  name_2[SH_MINIBUF] = '\0';
	}
      else
	{
	  i = strlen(name) - 1;
	  while (name[i] != '.' && i >= 0)
	    --i;
	  if (name[i] == '.') ++i;
	  strlcpy( name_2, &name[i], SH_MINIBUF +1 );
	}
	    
    }

  return name_2;
}

/* #ifndef SH_STANDALONE */
#endif

#ifndef USE_SRP_PROTOCOL

extern char TcpFlag[9];

void sh_passwd (char * salt, char * password, char * nounce, char *hash)
{

  char           *combi;
  long            len;
  register int i;
  unsigned char * dez = NULL;

  if (password == NULL)
    dez = (unsigned char *) &TcpFlag[0];
  else if (strlen(password) < 8)
    {
      fprintf(stderr, _("Password has less than 8 chars !\n"));
      _exit(EXIT_FAILURE);
    }

  if (password == NULL)
    {
      for (i = 0; i < 8; ++i)
	{ 
	  skey->vernam[i] = (*dez != '\0' ?  *dez : '\001'); 
	  ++dez; 
	}
      skey->vernam[8] = '\0';

      strlcpy (skey->vernam,
	       sh_tiger_hash(skey->vernam, TIGER_DATA, 8), KEY_MAX+1);
      skey->vernam[16] = skey->vernam[48]; 
      skey->vernam[33] = skey->vernam[49]; 
      skey->vernam[48] = '\0';
    }
  else
    {
      strlcpy (skey->vernam, password, KEY_MAX);
      skey->vernam[48] = '\0';
    }

  len = strlen(salt) + strlen(skey->vernam) + 1;
  if (nounce != NULL) len += strlen(nounce);
  
  /* H(s,P)
   */
  combi = SH_ALLOC(len);
  strlcpy (combi, salt, len);
  strlcat (combi, skey->vernam, len);
  if (nounce != NULL)
    strlcat (combi, nounce, len);
  strlcpy (hash, 
	   sh_tiger_hash(combi, TIGER_DATA, strlen(combi)),
	   KEY_MAX+1);

  /*
    fprintf(stderr, "DD: A: <%s>\n", salt);
    fprintf(stderr, "DD: P: <%s>\n", TcpFlag); 
    fprintf(stderr, "DD: V: <%s>\n", skey->vernam); 
    fprintf(stderr, "DD: C: <%s>\n", combi);
    fprintf(stderr, "DD: H: <%s>\n", hash); 
  */

  SH_FREE (combi);
  hash[16] = hash[48]; 
  hash[33] = hash[49]; 
  hash[48] = '\0';
  return;
}
#endif

#define SH_DO_WRITE 0
#define SH_DO_READ  1

#ifdef SH_WITH_CLIENT

static char * sh_logserver = NULL;

int sh_forward_setlogserver (char * address)
{
  if (address != NULL && sh_logserver == NULL) 
    {
      sh_logserver = (char *) malloc (strlen(address)+1);
      if (sh_logserver != NULL)
	{
	  strlcpy (sh_logserver, address, strlen(address)+1);
	  return GOOD;
	}
    }
  return BAD;
}


int sh_write_select(int type, int sockfd, char *buf, int nbytes, int * w_error)
{
  int    countbytes, count;
  fd_set fds;
  struct timeval tv;
  int    select_now;
  int    num_sel;
  
  struct  sigaction  new_act;
  struct  sigaction  old_act;

  /* ignore SIGPIPE (instead get EPIPE if connection is closed)
   */
  new_act.sa_handler = SIG_IGN;
  sigaction (SIGPIPE, &new_act, &old_act);
  
  FD_ZERO(&fds);
  FD_SET(sockfd, &fds);

  countbytes   = 0;
  tv.tv_sec    = 1;
  tv.tv_usec   = 0;
  select_now   = 0;

  while ( countbytes < nbytes ) {

    FD_ZERO(&fds);
    FD_SET(sockfd, &fds);

    if (type == SH_DO_WRITE) 
      {
	if ( (num_sel = select (sockfd+1, NULL, &fds, NULL, &tv)) == -1) 
	  {
	    if ( errno == EINTR ) /* try again */
	      continue;
	    *w_error = errno;
	    sigaction (SIGPIPE, &old_act, NULL);
	    return countbytes;
	  }
      }
    else
      {
	if ( (num_sel = select (sockfd+1, &fds, NULL, NULL, &tv)) == -1) 
	  {
	    if ( errno == EINTR ) /* try again */
	      continue;
	    *w_error = errno;
	    sigaction (SIGPIPE, &old_act, NULL);
	    return countbytes;
	  }
      }
      
    /* on Linux, timeout  is  modified to reflect the amount of
     * time not slept
     */
    tv.tv_sec    = 1;
    tv.tv_usec   = 0;

    /* let's not hang on forever
     */
    if (num_sel == 0) 
      {
	++select_now;       /* timeout */
	if ( select_now > (SH_SELECT_REPEAT * 5) )  /* 5 minutes */
	  {
	    *w_error = errno;
	    sh_error_handle(SH_ERR_ERR, PLACE
			    NULL,
			    _("Timeout"));
	    sigaction (SIGPIPE, &old_act, NULL);
	    return countbytes;
	  }
      }
    
    if ( FD_ISSET (sockfd, &fds) ) 
      {
	if (type == SH_DO_WRITE)
	  count = write (sockfd, buf, nbytes-countbytes);
	else
	  count = read  (sockfd, buf, nbytes-countbytes);

	if (count > 0) 
	{
	  countbytes += count;
	  buf        += count;    /* move buffer pointer forward */
	  if (countbytes < nbytes) FD_SET( sockfd, &fds );
	}
	else if (count < 0)
	  {
	    *w_error = errno;
	    sigaction (SIGPIPE, &old_act, NULL);
	    return countbytes;
	  }
	else /* count == 0 */
	  {
	    *w_error = errno;
	    sigaction (SIGPIPE, &old_act, NULL);
	    return countbytes;
	  }
      }
  }

  /* restore signal handler
   */
  sigaction (SIGPIPE, &old_act, NULL);

  return countbytes;
}

typedef struct {
  unsigned char head[SH_HEADER_SIZE];   /* protocol, siz_hi, siz_lo */ 
  char buf[SH_BUFSIZE+1];               /* the message to transmit  */
} sh_tcp_t;



int sh_forward_send(int socket, char * msgbuf, int length, char protocol)
{
  register int i;
  int          numbytes, countbytes;
  int          flag_err;
  sh_tcp_t     msg;

  if (length > SH_BUFSIZE) length = SH_BUFSIZE;
  for (i = 0; i < length; ++i) msg.buf[i] = msgbuf[i];

  msg.head[0] = protocol;
  msg.head[1] = (unsigned int)(length/256);
  msg.head[2] = (unsigned int)(length-256*msg.head[1]);
  msg.head[3] = 0x01;
  msg.head[4] = 0x01;
  msg.head[5] = 0x01;
  msg.head[6] = 0x01;
  

  numbytes     = SH_HEADER_SIZE;
  countbytes   = sh_write_select (SH_DO_WRITE, socket, 
				  msg.head, numbytes,
				  &flag_err);
  if (countbytes == numbytes)
    {
      numbytes     = (msg.head[1]*256 + msg.head[2]);
      countbytes   = sh_write_select (SH_DO_WRITE, socket, 
				      msg.buf, numbytes,
				      &flag_err);
    }

  if (countbytes == numbytes)
    return 0;
  else
    return flag_err;
}

/* receive answer, add a trailing NULL to terminate string
 */
int sh_forward_receive(int socket, char * msgbuf, int length, char * u)
{
  register int i = 0;
  int          numbytes, countbytes, wbytes;
  int          flag_err = 0;
  sh_tcp_t     msg;

  u[4]         = '\0';
  msgbuf[0]    = '\0';

  numbytes     = SH_HEADER_SIZE;
  countbytes   = sh_write_select (SH_DO_READ, socket, 
				  msg.head, numbytes,
				  &flag_err);
  if (countbytes != numbytes)
    {
      return flag_err;
    }
  else
    {
      numbytes     = (msg.head[1]*256 + msg.head[2]);
      u[0]         = msg.head[3];
      u[1]         = msg.head[4];
      u[2]         = msg.head[5];
      u[3]         = msg.head[6];
      u[4]         = '\0';
      numbytes = (numbytes > SH_BUFSIZE ? (SH_BUFSIZE+1) : numbytes);
      countbytes   = sh_write_select (SH_DO_READ, socket, 
				      msg.buf, numbytes,
				      &flag_err);
    }

  if (countbytes >= length) wbytes = length-1;
  else                      wbytes = countbytes;
  for (i = 0; i < wbytes; ++i) msgbuf[i] = msg.buf[i];
  msgbuf[wbytes] = '\0';

  if (countbytes == numbytes)
    return 0;
  else
    return flag_err;
}

/**************************************************
 *
 *
 *  C L I E N T  
 *
 *
 ***************************************************/


#include <time.h>

static SH_FIFO * fifo = NULL;

int sh_forward_try (char * errmsg);

int sh_forward (char * errmsg)
{
  static int           have_server = GOOD;
  int    status;
  char * popmsg;

  /* No log server available.
   */
  if (sh_logserver == NULL && have_server == GOOD)
    {
      have_server = BAD;
      sh_error_handle(SH_ERR_ERR, PLACE
		      NULL,
		      _("No log server given."));
      return BAD;
    }
  else if (have_server == BAD)
    {
      return BAD;
    }

  if (fifo == NULL)
    {
      fifo = SH_ALLOC(sizeof(SH_FIFO));
      fifo_init(fifo);
    }

  /* Check for messages on the queue, and send them first
   */
  while (NULL != (popmsg = pop_list(fifo)) )
    {
      status = sh_forward_try (popmsg);
      if (status != GOOD)
	{
	  push_tail_list (fifo, popmsg);    
	  push_list (fifo, errmsg);   
	  SH_FREE(popmsg);
	  return BAD;
	}
      SH_FREE(popmsg);
    }

  /* Now send the error message.
   */ 
  status = sh_forward_try (errmsg);
  if (status == BAD)
    {
      push_list (fifo, errmsg);    /* push message on stack */
      return BAD;
    }    

  return GOOD;  
}
  
int sh_forward_try (char * errmsg)
{
  int                  sockfd;
  struct in_addr       haddr;   /* host address from numeric                */
  struct hostent     * he = NULL;
  struct sockaddr_in   the_addr;
  int                  flag_err;
  char               * answer;
  int                  isnumeric = GOOD;
  static int           initialized = BAD;
  static int           conn_state  = GOOD;
  char                 hash[KEY_MAX+1];
  long                 len;
  char               * buffer;

  char   foo_u[5];
  char   foo_M1[KEY_MAX+1];

#ifdef USE_SRP_PROTOCOL
  char   u_real[SH_CHALLENGE_SIZE];
  char * foo_A;
  char * foo_Sc;
#else
  char   nounce[KEY_MAX+1];
#endif

  
  static unsigned long time_now  = 1200;
  static unsigned long time_last =    0;


  /* No message to transmit.
   */
  if (errmsg == NULL && initialized == GOOD)
    return GOOD;
  
  /* Connection in bad state.
   */
  if (initialized == BAD || conn_state == BAD)
    {
      /* retry bad attempt only after some time (5 minutes)
       */
      time_now  = (unsigned long) time (NULL);
      if ((time_now - time_last) < 60) 
	{
	  return BAD;
	}
    }
  time_last  = (unsigned long) time (NULL);

  /* get hostname
   */
#ifdef HAVE_INET_ATON
  if (0 == inet_aton(sh_logserver, &haddr))
#else
  if ((unsigned long)-1  == (haddr.s_addr = inet_addr(sh_logserver))  )    
#endif
    {
      isnumeric = BAD;
      if ((he=gethostbyname(sh_logserver)) == NULL) 
	{ 
	  conn_state = BAD;
	  sh_error_handle(SH_ERR_ERR, PLACE
			  _("Host name lookup failed."),
			  _("Host name lookup failed: %s"),
			  sh_forward_errmessage(h_errno));
	  return BAD;
	}
    }
  else
    isnumeric = GOOD;
  
  /* create a socket
   */
  if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) 
    {
      conn_state = BAD;
      flag_err = errno;
      sh_error_handle(SH_ERR_ERR, PLACE
		      _("Socket creation failed."),
		      _("Socket creation failed: %s."),
		      sh_error_message(flag_err));
      return BAD;
    }
  
  the_addr.sin_family = AF_INET;     
  the_addr.sin_port   = htons(SH_DEFAULT_PORT); 
  if (isnumeric == BAD)
    the_addr.sin_addr   = *((struct in_addr *)he->h_addr);
  else
    the_addr.sin_addr   = haddr;
  memset (&(the_addr.sin_zero),'\0', 8);     
  
  if (connect(sockfd, (struct sockaddr *)&the_addr,
	      sizeof(struct sockaddr)) == -1) 
    {
      conn_state = BAD;
      flag_err = errno;
      sh_error_handle(SH_ERR_FATAL, PLACE
		      _("Could not connect to log server."),
		      _("Could not connect to log server: %s."),
		      sh_error_message(flag_err));
      return BAD;
    }

  conn_state = GOOD;

  flag_err = 0;

  answer = SH_ALLOC(512);
  MLOCK(answer, 512);

  /*************************
   *
   *  initialization
   * 
   */


#ifndef USE_SRP_PROTOCOL

  /* challenge/response authentication
   */
  if (initialized == BAD)
    {
      strlcpy (answer, _("SALT "),   512);
      strlcat (answer, sh.localhost, 512);

      flag_err = sh_forward_send    (sockfd, answer,  strlen(answer), 
					 SH_PROTO_SRP);

      /* get nounce from server
       */
      flag_err = sh_forward_receive (sockfd, answer,  511, foo_u);  

      /* entry point for jump from message forward if session key must
       * be re-initialized
       */	 
 initBlock:

      strlcpy(nounce, answer, KEY_MAX+1);

      /* compute hash key H(s, P) ...
       */
      sh_passwd (answer, NULL, NULL, foo_M1);

      /* send it to server
       */
      flag_err = sh_forward_send (sockfd, foo_M1, 
				      KEY_MAX+1, SH_PROTO_SRP);
      if (flag_err == 0)
	flag_err = sh_forward_receive (sockfd, answer, 511, foo_u);
      if (flag_err == 0 &&
	  0 == strncmp (answer, foo_M1, strlen(foo_M1)))
	{
	  sh_passwd ("42", NULL, nounce, foo_M1);
	  /*
	  fprintf(stderr, "DD Client: A: <%s>\n", nounce);
	  fprintf(stderr, "DD Client: K: <%s>\n", 
		  foo_M1);
	  */
	  strlcpy (skey->session, foo_M1, KEY_MAX+1);
	  initialized = GOOD;
	}

      if (initialized == BAD)
	{
	  sh_error_handle(SH_ERR_ERR, PLACE
			  NULL,
			  _("Session key negotiation failed"));
	  memset(answer, '\0', sizeof(answer));
	  MUNLOCK(answer, 512);
	  SH_FREE(answer);
	  return BAD;
	}
      else
	{
	  sh_error_handle(SH_ERR_INFO, PLACE
			  NULL,
			  _("Session key negotiated successfully"));
	}
    }

#else

  /* use SRP protocol           */
  /* #include "srp_client.frag" */


  /* This is the SRP authenticated key exchange protocol.
   * Produces a session key skey->session.
   */
  if (initialized == BAD)
    {
	strlcpy (answer, _("SALT "),   512);
	strlcat (answer, sh.localhost, 512);
	flag_err = sh_forward_send    (sockfd, answer,  strlen(answer), 
					 SH_PROTO_SRP);
	flag_err = sh_forward_receive (sockfd, answer,  511, foo_u);  
	  
	/* entry point for jump from message forward if session key must
	 * be re-initialized
	 */	 
     initBlock:

      if (GOOD == sh_srp_init())
	{
	  sh_srp_x (answer, NULL);  /* x        password      */
	  sh_srp_make_a ();         /* a        random number */
	  foo_A = sh_srp_A();       /* g^a                    */

	  flag_err = sh_forward_send    (sockfd, foo_A, 
					 strlen(foo_A)+1, SH_PROTO_SRP); 
	  if (flag_err == 0)
	    flag_err = sh_forward_receive (sockfd, answer, 511, foo_u);  
	  
	  /*                     u        nounce        */
	  /*                     B        answer        */
	  /*                     S = (B-g^x)^(a+ux)     */
	  if (flag_err == 0 && GOOD == sh_srp_check_zero (answer))
	    {
	      strlcpy(u_real, sh_tiger_hash(foo_u, TIGER_DATA, 4), 
		      SH_CHALLENGE_SIZE);
	      foo_Sc = sh_srp_S_c (u_real,
				   answer);

	      /* now send H(A,B,H(Sc)) and check */
	      strlcpy (foo_M1, 
		       sh_srp_M (foo_A, 
				 answer, 
				 sh_tiger_hash(foo_Sc, 
					       TIGER_DATA, 
					       strlen(foo_Sc))),
		       KEY_MAX+1);

	      flag_err = sh_forward_send (sockfd, foo_M1, 
					  KEY_MAX+1, SH_PROTO_SRP);
	      if (flag_err == 0)
		flag_err = sh_forward_receive (sockfd, answer, 511, foo_u);
	      if (flag_err == 0 &&
		  0 == strncmp (answer,
			      sh_srp_M (foo_A,
					foo_M1,
					sh_tiger_hash(foo_Sc,
						      TIGER_DATA,
						      strlen(foo_Sc))),
				KEY_MAX+1)) 
		{
		  strlcpy (skey->session, 
			   sh_tiger_hash(foo_Sc, 
					 TIGER_DATA,
					 strlen(foo_Sc)),
			   KEY_MAX+1);
		  initialized = GOOD;
		}
	      SH_FREE(foo_Sc);
	      sh_srp_exit();
	    }
	  SH_FREE(foo_A);
	}

      if (initialized == BAD)
	{
	  sh_error_handle(SH_ERR_ERR, PLACE
			  NULL,
			  _("Session key negotiation failed"));
	  memset(answer, '\0', sizeof(answer));
	  MUNLOCK(answer, 512);
	  SH_FREE(answer);
	  return BAD;
	}
      else
	{
	  sh_error_handle(SH_ERR_INFO, PLACE
			  NULL,
			  _("Session key negotiated successfully"));
	}

    }

#endif

  /* no message, just session key negotiated
   */
  if (errmsg == NULL)
    {
      memset(answer, '\0', sizeof(answer));

      MUNLOCK(answer, 512);

      SH_FREE(answer);
      return GOOD;
    }
  
  /***********************************************
   *
   * send the message
   *
   */

  strlcpy (answer, _("HELO "),   512);
  strlcat (answer, sh.localhost, 512);
  flag_err = sh_forward_send    (sockfd, answer, strlen(answer), 
				 SH_PROTO_MSG);     /* say HELO       */
  if (flag_err == 0)
    { 
      flag_err = sh_forward_receive (sockfd, answer, 255, /* get challenge  */
				     foo_u);   

      if (flag_err == 0)
	{

	  /* re-negotiate key
	   */

	  if (strncmp(foo_u, "INIT", 4) == 0) 
	    {
	      initialized = BAD;
	      goto initBlock;
	    }

	  /* Hash(msg,challenge,sessionkey)  
	   */

	  /*
	    fprintf(stderr, "DD Client: B: <%s>\n", 
	    errmsg);
	    fprintf(stderr, "DD Client: C: <%s>\n", 
	    answer);
	    fprintf(stderr, "DD Client: K: <%s>\n", 
	    skey->session);
	  */

	  strlcat(answer, skey->session, 512);
	  if (strlen(answer) > KEY_MAX)
	    len    = strlen(errmsg) + strlen(answer) + 1;
	  else
	    len    = strlen(errmsg) + KEY_MAX + 1;
	  buffer = SH_ALLOC(len);

	  MLOCK(buffer, len);

	  strlcpy(buffer, errmsg, len);
	  strlcat(buffer, answer, len);
	  strlcpy(hash, 
		  sh_tiger_hash (buffer, TIGER_DATA, strlen(buffer)), 
		  KEY_MAX+1);
	  strlcpy(buffer, errmsg, len);
	  strlcat(buffer, hash,   len);
	  flag_err = sh_forward_send    (sockfd, buffer, 
					 strlen(buffer)+1, SH_PROTO_MSG);
	  memset(buffer, '\0', len);

	  MUNLOCK(buffer, len);

	  SH_FREE(buffer);

	}
    }
      
  close (sockfd);

  memset(answer, '\0', sizeof(answer));

  MUNLOCK(answer, 512);

  SH_FREE(answer);


#ifdef SH_ERROR_H  
  if (flag_err != 0)
    {
      conn_state = BAD;
      sh_error_handle(SH_ERR_ERR, PLACE
		      _("Connection error"),
		      _("Connection error: %s"),
		      sh_error_message(flag_err));
      return BAD;
    }
#endif

  return GOOD;
}

/* #ifdef SH_WITH_CLIENT */
#endif

#if defined (SH_WITH_CLIENT) || defined (SH_WITH_SERVER)

#ifndef USE_SRP_PROTOCOL

int sh_forward_make_client (char * str)
{
    char *          safer;
    char            key[KEY_MAX+1];
    unsigned char   in[9];
    int    i = 0, j, k, l = 0;

    if (strlen(str) != 16) 
      {
	fprintf(stderr, 
		_("Input must be a 16 digit hexadecimal number"\
		" (only 0-9, a-f, A-F allowed in input)\n"));
	_exit(EXIT_FAILURE);
      }

    while (i < 16)
      {
	k = sh_util_hexchar(str[i]); j = sh_util_hexchar(str[i+1]); 
	if (k != -1 && j != -1) 
	  {
	    in[l] = (k * 16 + j);
	    ++l; i+= 2;
	  }
	else
	  {
	    fprintf(stderr, _("Invalid char %c\n"), str[i]);
	    _exit(EXIT_FAILURE);
	  }
      }
    in[8] = '\0';

    for (i=0; i<8; ++i) 
      if (in[i] == '\0')  in[i] = '\001';
    
    safer = sh_util_safe_name (in);
    fprintf(stdout, _("Password |%s|\n"), safer);
    SH_FREE(safer);

    strlcpy (key, sh_tiger_hash (in, TIGER_DATA, 8), KEY_MAX+1);
    key[16] = key[48]; 
    key[33] = key[49]; 
    key[48] = '\0';

    fprintf(stdout, _("Client entry: Client=HOSTNAME@00000000@%s\n"), 
	    key);

    _exit(EXIT_SUCCESS);
    return 0;
}

#else

int sh_forward_make_client (char * str)
{
    char * foo_v;
    char * safer;
    char   salt[17];
    char   key[KEY_MAX+1];
    char   in[9];
    int    i = 0, j, k, l = 0;

    if (strlen(str) != 16) 
      {
	fprintf(stderr, 
		_("Input must be a 16 digit hexadecimal number"\
		" (only 0-9, a-f, A-F allowed in input)\n"));
	_exit(EXIT_FAILURE);
      }

    while (i < 16)
      {
	k = sh_util_hexchar(str[i]); j = sh_util_hexchar(str[i+1]); 
	if (k != -1 && j != -1) 
	  {
	    in[l] = (k * 16 + j);
	    ++l; i+= 2;
	  }
	else
	  {
	    fprintf(stderr, _("Invalid char %c\n"), str[i]);
	    _exit(EXIT_FAILURE);
	  }
      }
    in[8] = '\0';
    
    for (i=0; i<8; ++i) 
      if (in[i] == '\0')  in[i] = '\001';
    
    if (GOOD == sh_srp_init())
      {
	safer = sh_util_safe_name (in);
	fprintf(stdout, _("Password |%s|\n"), safer);
	SH_FREE(safer);
	sh_util_keyinit(key, KEY_MAX);
	strlcpy(salt, sh_tiger_hash(key, TIGER_DATA, KEY_MAX), 17); 
	sh_srp_x (salt, in);
	foo_v  = sh_srp_verifier ();
	fprintf(stdout, _("Client entry: Client=HOSTNAME@%s@%s\n"), 
		salt, foo_v);
	SH_FREE(foo_v);
	sh_srp_exit();
	_exit(EXIT_SUCCESS);
      }
    _exit(EXIT_FAILURE);
    return -1;
}
#endif

#endif
  
/**************************************************
 *
 *
 *  S E R V E R   
 *
 *
 ***************************************************/

#ifdef SH_WITH_SERVER

#define CONN_FREE    0
#define CONN_READING 1
#define CONN_SENDING 2
#define CONN_PAUSE   3
#define CONN_BUSY    4

#include <time.h>


typedef struct client_entry {
  char                  * hostname;
  char                  * salt;
  char                  * verifier;
  char                    session_key[KEY_MAX+1];
  struct client_entry   * left;
  struct client_entry   * right;
  struct client_entry   * up;
} client_t;

static client_t * all_clients = NULL;


int sh_forward_register_client (char * str)
{
  client_t   * newclt;
  client_t   * here;
  client_t   * up;

  char       * ptr;
  int          sepnum = 0;
  int          sep[2];
  register int i = 0;
  int          treetest;

  ptr = str; 
  while (*ptr) {
    if (*ptr == '@' && sepnum < 2 ) 
      { 
	sep[sepnum] = i;
	++sepnum;
      } 
    ++ptr; ++i; 
  }

  
  if (sepnum == 2)
    {
      newclt = SH_ALLOC (sizeof(client_t));
      newclt->hostname = SH_ALLOC (sep[0]+1);
      newclt->salt     = SH_ALLOC (sep[1]-sep[0]);
      newclt->verifier = SH_ALLOC (strlen(str)-sep[1]+1);
      newclt->session_key[0]   = '\0';
      strlcpy(newclt->hostname, &str[0],            sep[0]+1);
      strlcpy(newclt->salt,     &str[sep[0]+1],     sep[1]-sep[0]);
      strlcpy(newclt->verifier, &str[sep[1]+1],     strlen(str)-sep[1]+1);

      /* insert in binary search tree
       */
      newclt->left     = NULL;
      newclt->right    = NULL;
      newclt->up       = NULL;
      here = all_clients;
      up   = NULL;
      while (here) 
	{
	  treetest = strcmp (newclt->hostname, here->hostname);
	  if (treetest == 0)
	    { 
	      SH_FREE(newclt->hostname);
	      SH_FREE(newclt->salt);
	      SH_FREE(newclt->verifier);
	      SH_FREE(newclt);
	      return (BAD);
	    }
	  up   = here;
	  here = (treetest < 0) ? here->left : here->right;
	}
      newclt->up = up;
      
      if(up)
	{
	  if(strcmp (newclt->hostname, up->hostname) < 0)
	    up->left  = newclt;
	  else
	    up->right = newclt;
	}
      else
	all_clients = newclt;


      /******************************
      if (all_clients == NULL)
	newclt->next     = NULL;
      else
	newclt->next     = all_clients;
      all_clients      = newclt;
      ******************************/

      sh_error_handle(SH_ERR_ALL, PLACE
		      NULL,
		      _("Registered %s, salt %s, verifier %s"),
		      newclt->hostname,
		      newclt->salt,
		      newclt->verifier);
    }
  else
    {
      sh_error_handle(SH_ERR_ERR, PLACE
		      NULL,
		      _("Bad client registry entry %s"),
		      str);
      return BAD;
    } 
      
  return GOOD;
}



typedef struct {
  int             state;
  int             fd;
  char          * buf;
  unsigned char   head[SH_HEADER_SIZE];
  char            challenge[SH_CHALLENGE_SIZE];
  char            peer[SH_MINIBUF+1];
  client_t      * client_entry;
  char          * K;
  char          * M1;
  char          * A;
  int             headcount;
  unsigned long   bytecount;
  unsigned long   bytes_to_send;
  unsigned long   bytes_to_get;
  int             pass;
  unsigned long   timer;
} sh_conn_t;



void sh_forward_do_free (sh_conn_t * conn)
{
  if (conn->K != NULL) 
    {
      SH_FREE(conn->K);
      conn->K           = NULL;
    }
  if (conn->A != NULL) 
    {
      SH_FREE(conn->A);
      conn->A           = NULL;
    }
  if (conn->M1 != NULL) 
    {
      SH_FREE(conn->M1);
      conn->M1           = NULL;
    }
  if (conn->buf != NULL) 
    {
      SH_FREE(conn->buf);
      conn->buf          = NULL;
    }
  close (conn->fd);
  conn->fd            = -1;
  memset(conn->challenge, '\0', SH_CHALLENGE_SIZE);
  conn->state         = CONN_FREE;
  conn->headcount     = 0;
  conn->bytecount     = 0;
  conn->bytes_to_send = 0;
  conn->bytes_to_get  = 0;
  conn->pass          = 0;
  conn->timer         = 0;
  conn->client_entry  = NULL;
  return;
}

void sh_forward_prep_send (sh_conn_t * conn, 
			   char * msg, int length,
			   char * u, char protocol)
{
  register int i;

  conn->headcount     = 0;
  conn->bytecount     = 0;
  conn->bytes_to_send = 0;
  conn->bytes_to_get  = 0;
  if (conn->buf != NULL) 
    {
      SH_FREE(conn->buf);
      conn->buf           = NULL;
    }
  conn->buf       = SH_ALLOC(length+1);
  for (i = 0; i < length; ++i) conn->buf[i] = msg[i];
  conn->buf[length] = '\0';
  conn->head[0]   = protocol;
  conn->head[1]   = (unsigned int)(length/256);
  conn->head[2]   = (unsigned int)(length-256*conn->head[1]);
  conn->head[3]   = u[0];
  conn->head[4]   = u[1];
  conn->head[5]   = u[2];
  conn->head[6]   = u[3];
  conn->state     = CONN_SENDING;
  return;
}

#include <sys/times.h>

/* here we check the message received, and decide on the answer to send
 * (if any). The connection is in CONN_PAUSED state, thus we must:
 * (i)   define the proper reaction
 * (ii)  reset to CONN_READING or CONN_WRITING or CONN_FREE
 * (iii) eventually reset the connection entry
 */
void check_protocol(sh_conn_t * conn, int state)
{
  client_t         * this_client; 
  int                treetest;

  char       hash[KEY_MAX+1];
  char     * buffer;
  long       len;
  struct tms ttx;
  unsigned long ticks;
  int        i;
  char     * test;
  char       u[5] = "OOOO";

#ifdef USE_SRP_PROTOCOL
  char     * foo_B;
  char     * foo_Ss;
#endif

  /* protocols: 
   * -- (ii)     authenticated message transfer
   * -- (i)      SRP key exchange
   */

  /* ---------  message exchange  -----------
   */
  if ( conn->head[0] == SH_PROTO_MSG )
    {

      if (state == SH_DO_READ)        /* finished reading */
	{

	  /* client requests challenge
	   */
	  if (strncmp(conn->buf, _("HELO"), 4) == 0)
	    {

	      /* search the register
	       */
	      this_client = all_clients;
	      while (this_client)
		{
		  treetest = strncmp (&(conn->buf[5]), this_client->hostname, 
				      strlen(this_client->hostname));
		  if (treetest == 0)
		    break;
		  if (treetest < 0)
		    this_client = this_client->left;
		  else
		    this_client = this_client->right;
		}

	      if (this_client == NULL)
		{
		  sh_error_handle(SH_ERR_ERR, PLACE
				  NULL,
				  _("Connection attempt "\
				    "from unregistered host %s"),
				  &(conn->buf[5]));
		  sh_forward_do_free (conn);
		  return;
		} 

	      if (this_client->session_key[0] == '\0')
		{
		  /* fake an auth request and jump there
		   */
		  conn->head[0] = SH_PROTO_SRP;
		  conn->buf[0] = 'S';
		  conn->buf[1] = 'A';
		  conn->buf[2] = 'L';
		  conn->buf[3] = 'T';
		  goto servInit;
		}
	      
	      conn->client_entry = this_client;
	      strlcpy (conn->peer, &(conn->buf[5]), SH_MINIBUF+1);

	      ticks = (unsigned long) times (&ttx);
	      test = (char *) &ticks;
	      sh_util_cpylong (conn->challenge, test, 4);
	      conn->challenge[4] = '\0';
	      for (i = 0; i < 5; ++i) 
		u[i] = conn->challenge[i];
	      sh_forward_prep_send (conn, conn->challenge, 5, u, 
				    SH_PROTO_MSG);
	    }

	  /* client has send a message
	   */
	  else if ( (len = strlen(conn->buf) - KEY_MAX) > 0)
	    {
	      /* get hash from message end, truncate message
	       */
	      strlcpy(hash, &(conn->buf[len]), KEY_MAX+1);
	      conn->buf[len] = '\0';
	      
	      /* verify hash
	       */
	      len    = strlen(conn->buf) +
		strlen(conn->client_entry->session_key) + KEY_MAX + 6;
	      buffer = SH_ALLOC(len);
	      strlcpy(buffer, conn->buf,                          len);
	      strlcat(buffer, conn->challenge,                    len);
	      strlcat(buffer, conn->client_entry->session_key,    len);

	      /*
	      fprintf(stderr, "DD Server: B: <%s>\n", 
		      conn->buf);
	      fprintf(stderr, "DD Server: C: <%s>\n", 
		      conn->challenge);
	      fprintf(stderr, "DD Server: K: <%s>\n", 
		      conn->client_entry->session_key);
	      */

	      if (strncmp(hash, sh_tiger_hash (buffer, 
					       TIGER_DATA, 
					       strlen(buffer)),
			  KEY_MAX) != 0)
		sh_error_handle(SH_ERR_ERR, PLACE
				NULL,
				_("Bogus connection from %s"),
				conn->peer);
	      else
		sh_error_handle(SH_ERR_INET, PLACE
				NULL,
				_("<%s> %s"),
				sh_strip_domain (conn->peer),
				conn->buf);
	      SH_FREE(buffer);
	      sh_forward_do_free (conn);
	      
	    }
	  
	  /* client does something unexpected
	   */
	  else  /* ---- ??? ----- */
	    {
	      sh_forward_do_free (conn);
	    }
	}
      else if (state == SH_DO_WRITE)  /* finished writing */
	{
	  /* challenge is sent, now wait for message from client
	   */
	  conn->headcount     = 0;
	  conn->bytecount     = 0;
	  conn->bytes_to_send = 0;
	  conn->bytes_to_get  = 0;
	  if (conn->buf != NULL) 
	    {
	      SH_FREE(conn->buf);
	      conn->buf           = NULL;
	    }
	  conn->state     = CONN_READING;
	}
      return;
    }

  /* ---------  authentication  -----------
   */

  /* entry point for jump from message forward if session key must
   * be re-initialized
   */	 
 servInit:

  if ( conn->head[0] == SH_PROTO_SRP )
    {

#ifndef USE_SRP_PROTOCOL

      if (state == SH_DO_READ)        /* finished reading */
	{
	  /* first pass -- client request salt  
	   */
	  if (strncmp(conn->buf, _("SALT"), 4) == 0)
	    {

	      /* search the register
	       */
	      this_client = all_clients;
	      while (this_client)
		{
		  treetest = strncmp (&(conn->buf[5]), this_client->hostname, 
				      strlen(this_client->hostname));
		  if (treetest == 0)
		    break;
		  if (treetest < 0)
		    this_client = this_client->left;
		  else
		    this_client = this_client->right;
		}

	      if (this_client == NULL)
		{
		  sh_error_handle(SH_ERR_ERR, PLACE
				  NULL,
				  _("Connection attempt "\
				    "from unregistered host %s"),
				  &(conn->buf[5]));
		  sh_forward_do_free (conn);
		  return;
		} 

	      
	      conn->client_entry = this_client;
	      strlcpy (conn->peer, &(conn->buf[5]), SH_MINIBUF+1);

	      /* 
	       * create nounce u
	       */
	      ticks = times (&ttx);          /* u        nounce        */
	      test = (char *) &ticks;
	      sh_util_cpylong (u, test, 4);
	      u[4] = '\0';
	      conn->A = SH_ALLOC(KEY_MAX+1);
	      strlcpy(conn->A, 
		      sh_tiger_hash(u, TIGER_DATA, 4),
		      KEY_MAX+1);
	      u[0] = 'I'; u[1] = 'N'; u[2] = 'I'; u[3] = 'T';

      	      sh_forward_prep_send (conn, 
				    conn->A, 
				    strlen(conn->A), 
				    u, 
				    SH_PROTO_SRP);
	    }

	  /* client -- third pass
	   * A := H(nounce), verifier := H(password), 
	   */
	  else if (conn->pass == 3)
	    {

	      conn->K = SH_ALLOC(KEY_MAX+1);

	      /* compute hash key H(s, P) ...
	       */
	      sh_passwd (conn->A, conn->client_entry->verifier, NULL, conn->K);

	      /*
		fprintf(stderr, "DD Server: A: <%s>\n", conn->A);
		fprintf(stderr, "DD Server: V: <%s>\n", 
		conn->client_entry->verifier);
		fprintf(stderr, "DD Server: K: <%s>\n", 
		conn->K);
		fprintf(stderr, "DD Server: Kc: <%s>\n", 
		      conn->buf);
	      */

	      if (0 == strncmp(conn->K, conn->buf, KEY_MAX-2))
		{
		  sh_forward_prep_send (conn, 
					conn->K, 
					strlen(conn->K), 
					u, 
					SH_PROTO_SRP);
		  sh_error_handle(SH_ERR_SEVERE, PLACE
				  NULL,
				  _("Session key for %s "\
				    "successfully negotiated"),
				  conn->peer);
		  /* session key
		   */
		  sh_passwd ("42", conn->client_entry->verifier, 
			     conn->A, conn->K);
		  strlcpy(conn->client_entry->session_key, conn->K, KEY_MAX+1);
		  SH_FREE (conn->K);
		  conn->K = NULL;
		  SH_FREE (conn->A);
		  conn->A = NULL;
		}
	      else
		{
		  sh_forward_prep_send (conn, 
					"BAD", 
					3, 
					u, 
					SH_PROTO_SRP);
		  sh_error_handle(SH_ERR_ERR, PLACE
				  NULL,
				  _("Could not verify session key - host %s"),
				  conn->peer);
		  sh_forward_do_free (conn);
		} 
	    }
	}

#else
      /* use SRP                    */
      /* #include "srp_server.frag" */


      if (state == SH_DO_READ)        /* finished reading */
	{

	  /* first pass -- client request salt  
	   */
	  if (strncmp(conn->buf, _("SALT"), 4) == 0)
	    {
	      
	      /* search the register
	       */
	      this_client = all_clients;
	      while (this_client)
		{
		  treetest = strncmp (&(conn->buf[5]), 
				      this_client->hostname, 
				      strlen(this_client->hostname));
		  if (treetest == 0)
		    break;
		  if (treetest < 0)
		    this_client = this_client->left;
		  else
		    this_client = this_client->right;
		}
	      
	      if (this_client == NULL)
		{
		  sh_error_handle(SH_ERR_ERR, PLACE
				  NULL,
				  _("Connection attempt "\
				    "from unregistered host %s"),
				  &(conn->buf[5]));
		  sh_forward_do_free (conn);
		  return;
		} 
	      
	      
	      conn->client_entry = this_client;
	      strlcpy (conn->peer, &(conn->buf[5]), SH_MINIBUF+1);

	      u[0] = 'I'; u[1] = 'N'; u[2] = 'I'; u[3] = 'T'; u[4] = '\0';
	      
	      sh_forward_prep_send (conn, 
				    conn->client_entry->salt, 
				    strlen(conn->client_entry->salt), 
				    u, 
				    SH_PROTO_SRP);
	    }

	  /* client has sent A -- third pass
	   */
	  else if (conn->pass == 3 && GOOD == sh_srp_init())
	    {
	      /* check A, only send B if correct 
	       */
	      
	      if ( strlen(conn->buf) < SH_BUFSIZE && 
		   GOOD == sh_srp_check_zero (conn->buf) )
		{
		  len = strlen(conn->buf)+1;
		  conn->A = SH_ALLOC(len);
		  strlcpy (conn->A, conn->buf, len);
		  
		  /* 
		   * compute B 
		   */
		  sh_srp_make_a ();              /* b        random number */
		  foo_B = sh_srp_B               /* B = v + g^b            */
		    (conn->client_entry->verifier);

		  /* 
		   * create nounce u
		   */
		  ticks = times (&ttx);          /* u        nounce        */
		  test = (char *) &ticks;
		  sh_util_cpylong (u, test, 4);
		  u[4] = '\0';
		  strlcpy(conn->challenge, 
			  sh_tiger_hash(u, TIGER_DATA, 4),
			  SH_CHALLENGE_SIZE);

		  /* 
		   * compute the session key K and M1 = Hash(A,B,K)
		   */
		  foo_Ss = sh_srp_S_s (conn->challenge, 
				       conn->A, 
				       conn->client_entry->verifier);
		  conn->K = SH_ALLOC(KEY_MAX+1);
		  strlcpy(conn->K, 
			  sh_tiger_hash( foo_Ss, TIGER_DATA, strlen(foo_Ss)),
			  KEY_MAX+1);
		  conn->M1 = SH_ALLOC(KEY_MAX+1);
		  strlcpy(conn->M1, 
			  sh_srp_M (conn->A, foo_B, conn->K),
			  KEY_MAX+1);
		  /*
		   * send B
		   */
		  sh_forward_prep_send (conn, 
					foo_B,
					strlen(foo_B)+1,
					u,
					SH_PROTO_SRP);
		  SH_FREE(foo_Ss);
		  SH_FREE(foo_B);
		}
	      else
		{
		  sh_error_handle(SH_ERR_ERR, PLACE
				  NULL,
				  _("Could not verify public key A - host %s"),
				  conn->peer);
		  sh_forward_do_free (conn);
		}
	      sh_srp_exit();
	    }
		  
	  /* client has sent M1 -- fifth pass
	   */
	  else if (conn->pass == 5)
	    {
	      /* check that we agree on M1
	       */
	      if (strncmp(conn->buf, conn->M1, KEY_MAX) == 0)
		{
		  /*
		   * send M2
		   */
		  sh_forward_prep_send (conn, 
					sh_srp_M (conn->A, conn->M1, conn->K),
					KEY_MAX+1,
					u,
					SH_PROTO_SRP);
		  SH_FREE(conn->A);  conn->A  = NULL;
		  SH_FREE(conn->M1); conn->M1 = NULL;
		  strlcpy(conn->client_entry->session_key, conn->K, KEY_MAX+1);
		  SH_FREE(conn->K);  conn->K  = NULL;
		  sh_error_handle(SH_ERR_SEVERE, PLACE
				  NULL,
				  _("Session key for %s "\
				    "successfully negotiated"),
				  conn->peer);
		}
	      else
		{
		  sh_error_handle(SH_ERR_ERR, PLACE
				  NULL,
				  _("Could not verify session key - host %s"),
				  conn->peer);
		  sh_forward_do_free (conn);
		} 
	    }
	}

#endif

      else if (state == SH_DO_WRITE)  /* finished writing */
	{
	  conn->headcount     = 0;
	  conn->bytecount     = 0;
	  conn->bytes_to_send = 0;
	  conn->bytes_to_get  = 0;
	  if (conn->buf != NULL) 
	    {
	      SH_FREE(conn->buf);
	      conn->buf           = NULL;
	    }
	  conn->state     = CONN_READING;
	}
    }
}


int sh_forward_do_read (sh_conn_t * conn)
{
  int   byteread;     /* bytes read         */

  if (conn->headcount < SH_HEADER_SIZE) 
    {
      conn->bytes_to_get = SH_HEADER_SIZE - conn->headcount;
      byteread           = read (conn->fd, &(conn->head[conn->headcount]),
				 conn->bytes_to_get);
      if (byteread > 0) 
	{
	  conn->headcount += byteread;
	  if (conn->headcount == SH_HEADER_SIZE)
	    conn->bytes_to_get = 
	      (256 * (int)conn->head[1] + (int)conn->head[2]);
	}
      else
	{
	  sh_forward_do_free ( conn );
	  return BAD;
	}
      return GOOD;
    }

  if (conn->headcount == SH_HEADER_SIZE && conn->bytes_to_get > 0)
    {
      if (conn->buf == NULL) 
	{
	  conn->buf       = SH_ALLOC (conn->bytes_to_get + 1);
	  conn->bytecount = 0;
	}
      byteread           = read (conn->fd, &(conn->buf[conn->bytecount]),
				 conn->bytes_to_get - conn->bytecount);
      if (byteread > 0) 
	{
	  conn->bytecount    += byteread;
	  if (conn->bytecount == conn->bytes_to_get) 
	    {
	      ++conn->pass;
	      /* always terminate with NULL - we might use strcmp()
	       */
	      conn->buf[conn->bytecount] = '\0';
	      conn->state                = CONN_PAUSE;
	      /* ------  HERE CALL check_protocol(conn) -------  */
	      check_protocol(conn, SH_DO_READ);
	    }
	}
      else
	{
	  sh_forward_do_free ( conn );
	  return BAD;
	}
    }
  return GOOD;
}

int sh_forward_do_write (sh_conn_t * conn)
{
  int   bytesent;     /* bytes read         */

  if (conn->headcount < SH_HEADER_SIZE) 
    {
      conn->bytes_to_send = SH_HEADER_SIZE - conn->headcount;
      bytesent            = write (conn->fd, 
				   &(conn->head[conn->headcount]), 
				   conn->bytes_to_send);
      if (bytesent > 0) 
	{
	  conn->headcount += bytesent;
	  if (conn->headcount == SH_HEADER_SIZE) 
	    conn->bytes_to_send = 
	      (256 * (int)conn->head[1] + (int)conn->head[2]);
	}
      else 
	{
	  sh_forward_do_free ( conn );
	  return BAD;
	}
      return GOOD;
    }

  if (conn->headcount == SH_HEADER_SIZE && conn->bytes_to_send > 0)
    {
      bytesent           = write (conn->fd, &(conn->buf[conn->bytecount]), 
				  conn->bytes_to_send - conn->bytecount);
      if (bytesent > 0) 
	{
	  conn->bytecount    += bytesent;
	  if (conn->bytecount == conn->bytes_to_send) 
	    {
	      ++conn->pass;
	      conn->state         = CONN_PAUSE;
	      /* ------  HERE CALL check_protocol(conn) -------  */
	      check_protocol(conn, SH_DO_WRITE);

	    }
	}
      else
	{
	  sh_forward_do_free ( conn );
	  return BAD;
	}
    }
  return GOOD;
}

 
int sh_forward_accept (int sock, sh_conn_t * newconn)
{
  int                rc;
  struct sockaddr_in addr;
  int                addrlen = sizeof(addr);
  int                errflag;

  if ( (rc = accept(sock, (struct sockaddr *) &addr, &addrlen)) >= 0)
    {

      /* prepare for usage of connection
       */
      (void) fcntl( rc, F_SETFD, 1 );
      newconn->fd           = rc;
      newconn->state        = CONN_READING;
      newconn->timer        = (unsigned long) time (NULL);

      sh_error_handle(SH_ERR_INFO, PLACE
		      _("New connection, using file descriptor %d"),
		      _("New connection"),
		      rc);
      return GOOD;
    }
  else
    {
      errflag = errno;
      sh_error_handle(SH_ERR_ERR, PLACE
		      _("Connection error"),
		      _("Connection error: accept: %s"),
		      sh_error_message(errflag));

      newconn->fd    = -1;
      newconn->state = CONN_FREE;
      return BAD;
    }

}

void sh_forward_printerr(char * str, int errnum)
{
  sh_error_handle(SH_ERR_ERR, PLACE
		  _("Server error"),
		  _("Server error: %s: %s"),
		  str,
		  sh_error_message(errnum));
  return;
}

static    sh_conn_t        * conns = NULL;

void sh_receive()
{
    struct sockaddr_in addr;
    int addrlen      = sizeof(addr);

    int                sock;
    int                flag = 1;  /* non-zero to enable an option */

    sh_conn_t        * cx;
    fd_set             readset;
    fd_set             writeset;
    struct timeval     tv;
    int                num_sel;
    int                errnum;
    int                maxconn;
    int                nowconn;
    int                status;
    int                high_fd;
    register int       i;
    unsigned long      time_now;

    struct  sigaction  new_act;
    struct  sigaction  old_act;

    /* ignore SIGPIPE (instead get EPIPE if connection is closed)
     *      --- we have called sh_unix_init() already ---
     */
    new_act.sa_handler = SIG_IGN;
    sigaction (SIGPIPE, &new_act, &old_act);

    /* create the socket, bind() it and listen()
     */
    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
      {
	errnum = errno; 
	sh_forward_printerr (_("socket"), errnum);
	exit(EXIT_FAILURE);
      }
    (void) fcntl( sock, F_SETFD, 1 );

    if ( setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
			  (void *) &flag, sizeof(flag)) < 0 )
      {
	errnum = errno;
	sh_forward_printerr (_("setsockopt"), errnum);
	exit(EXIT_FAILURE);
      }

    addr.sin_family = AF_INET;
    addr.sin_port = htons(SH_DEFAULT_PORT);
    addr.sin_addr.s_addr = INADDR_ANY;

    if ( bind(sock, (struct sockaddr *) &addr, addrlen) < 0) 
      {
	errnum = errno;
	sh_forward_printerr (_("bind"), errnum);
	exit(EXIT_FAILURE);
      }

    if ( fcntl( sock, F_SETFL, O_NDELAY ) < 0 )
      {
	errnum = errno;
	sh_forward_printerr (_("fcntl"), errnum);
	exit(EXIT_FAILURE);
      }
	
    if ( listen(sock, 5) < 0)
      {
	errnum = errno;
	sh_forward_printerr (_("listen"), errnum);
	exit(EXIT_FAILURE);
      }

    sh_error_handle(SH_ERR_STAMP, PLACE
		    _("Server listening."),
		    _("Server listening on fd %d"),
		    sock);
 
    /* ****************************************************************
     *
     * This is a non-forking server. We use select() on the listen()
     * socket to watch for new connections. For new connections, accept()
     * will return a new socket that is put in the read/write filesets.
     * Data about active connections are kept in the 'conns' table. 
     *
     ******************************************************************/

    /* The table to hold info on sockets.
     */
    conns = SH_ALLOC (sizeof(sh_conn_t) * FD_SETSIZE);
    maxconn     = FD_SETSIZE;
    for (i = 1; i < maxconn; ++i)
      {
	conns[i].buf   = NULL;
	conns[i].K     = NULL;
	conns[i].A     = NULL;
	conns[i].M1    = NULL;
	sh_forward_do_free ( &conns[i]);
      }

    nowconn    = 1;
    tv.tv_sec  = 5;
    tv.tv_usec = 0;

    /* conns[0] is the listen() socket. Always in read mode.
     */
    conns[0].fd    = sock;
    conns[0].state = CONN_READING;

    high_fd = sock;

    /* This is the select() loop.
     */
    while (1 == 1)
      {

	/* Recompute the descriptor set. select() modifies it,
	 * thus we update it using the info from the connection table.
	 */
	FD_ZERO( &readset );
	FD_ZERO( &writeset );
	FD_SET(conns[0].fd, &readset );
	high_fd = conns[0].fd;
	time_now  = (unsigned long) time (NULL);
	for (i = 1; i < maxconn; ++i)
	  {
	    /* eliminate timed out connections
	     */
	    if (conns[i].state != CONN_FREE && time_now - conns[i].timer > 300)
	      {
		sh_error_handle(SH_ERR_ERR, PLACE
				NULL,
				_("Connection to %s timed out"),
				conns[i].peer);
		sh_forward_do_free ( &conns[i]);
	      }
	    if       (conns[i].state   == CONN_READING)
	      { 
		FD_SET(conns[i].fd, &readset);
		high_fd = (high_fd < conns[i].fd ? conns[i].fd : high_fd);
	      }
	    else if  (conns[i].state   == CONN_SENDING)
	      {
		FD_SET(conns[i].fd, &writeset);
		high_fd = (high_fd < conns[i].fd ? conns[i].fd : high_fd);
	      }
	  }

	/* Do the select().
	 */
	num_sel = select(high_fd+1, &readset, &writeset, NULL, &tv);

	/* reset timeout - modified by select() on some systems
	 */
	tv.tv_sec  = 5;
	tv.tv_usec = 0;

	/* Error handling.
	 */
	if ( num_sel < 0 )        /* some error             */
	  {
	    if ( errno == EINTR )
		continue;	  /* try again              */
	    errnum = errno;
	    sh_error_handle(SH_ERR_FATAL, PLACE
			    _("Fatal server error - select() %s"),
			    _("Error in select(): %s"),
			    sh_error_message(errnum));
	    exit( 1 );
	    }

	/* select() timeout handling.
	 */
	if ( num_sel == 0 )       /* timeout - no connection */ 
	  {
	    continue;
	  }

	/* New connection.
	 */
	if ( FD_ISSET(conns[0].fd , &readset )) /* a new connection   */
	  {
	    --num_sel;
	    status = GOOD;
	    if (nowconn < maxconn)
	      {
		i = 1;
		while (i < maxconn)
		  {
		    if (conns[i].state == CONN_FREE)
		      {
			status = sh_forward_accept (conns[0].fd, &conns[i]);
			if (status == GOOD) 
			  high_fd = 
			    (high_fd > conns[i].fd ? high_fd : conns[i].fd);
			break;
		      }
		    ++i;
		  }
	      }
	    if (status == GOOD) continue;
	  }

	/* Check for pending read/write on the rest of the sockets.
	 */
	for ( i = 0; num_sel > 0 && i < maxconn; ++i )
	  {
	    cx = &conns[i];
	    if ( cx->state == CONN_READING &&
		 FD_ISSET( cx->fd, &readset ) )
	      {
		--num_sel;
		sh_forward_do_read ( cx );
	      }
	    else if ( cx->state == CONN_SENDING &&
		 FD_ISSET( cx->fd, &writeset ) )
	      {
		--num_sel;
		sh_forward_do_write ( cx );
	      }
	  }

      }
}

void  free_client_tree (client_t * inptr)
{
  if (inptr == NULL)
    return;

  /* free left and right tree
   */
  free_client_tree (inptr->left);
  free_client_tree (inptr->right);

  /* free this node
   */
  if (inptr->hostname) SH_FREE(inptr->hostname);
  if (inptr->salt)     SH_FREE(inptr->salt);
  if (inptr->verifier) SH_FREE(inptr->verifier);
  inptr->left  = NULL;
  inptr->right = NULL;
  inptr->up    = NULL;
  SH_FREE(inptr);
  inptr        = NULL;

  return;
}

void sh_forward_free_all ()
{
  int maxconn     = FD_SETSIZE;
  register int i;
  
  if (conns != NULL)
    for (i = 1; i < maxconn; ++i)
      {
	sh_forward_do_free ( &conns[i]);
      }
  free_client_tree (all_clients);

  if (conns != NULL)
    SH_FREE (conns);

  return;
}
/* #ifdef SH_WITH_SERVER */
#endif


  
