#include <stdio.h>

/* #include <stdlib.h> */

#include "config_xor.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif

#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
#include <sys/mman.h>
#endif

#include "sh_tiger.h"

#include "sh_unix.h"
#include "sh_error.h"

typedef unsigned long sh_word32;
typedef unsigned char sh_byte;

void
tiger(sh_word32 *str, sh_word32 length, sh_word32 res[6]);

void
tiger_t(sh_word32 *str, sh_word32 length, sh_word32 res[6]);

static void tiger_dbg(sh_word32 res[6], int step, 
unsigned long nblocks, unsigned long ncount)
{
  /*
    fprintf(stderr,                                     
            _("ST %d BLK %2ld CT %2ld %08lX %08lX %08lX %08lX %08lX %08lX\n"),
	    step,
	    nblocks,
	    ncount,
            (sh_word32)(res[1]), 
            (sh_word32)(res[0]), 
            (sh_word32)(res[3]), 
            (sh_word32)(res[2]), 
            (sh_word32)(res[5]), 
            (sh_word32)(res[4]) );
  */
}

/* this is the wrapper function -- not part of the tiger reference
 * implementation
 */
char * sh_tiger_hash (char * filename, TigerType what, unsigned long Length)
{
  int  fd;
  int  i, j;
  int  count;
  int  blk;
  sh_byte * bptr;
  static char out[51];
  sh_byte buffer[PRIV_MAX+2];
  sh_byte bbuf[64];

  unsigned long ncount = 0, nblocks = 0;
  unsigned long  t, msb, lsb;

  sh_word32 res[6];


  /* initialize 
   */
  res[0]=0x89ABCDEF;
  res[1]=0x01234567;
  res[2]=0x76543210;
  res[3]=0xFEDCBA98;
  res[4]=0xC3B2E187;
  res[5]=0xF096A5B4;

  if (what == TIGER_FILE) {
    fd = sh_unix_fileopen (filename, 'r', 
                           SH_ISDATA, ShDFLevel[SH_ERR_T_FILE], YESPRIV);
    if (fd == (-1)) 
      return _("0000000000000000 000000000000000 000000000000000");

#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
    if (skey->mlock_failed == GOOD) 
      {
	sh_set_suid ();
        if ( (-1) == mlock( buffer, (PRIV_MAX+2) * sizeof(sh_byte)) ) 
          skey->mlock_failed = BAD;
	sh_unset_suid ();
      }
#else
    if (skey->mlock_failed == GOOD) {
      skey->mlock_failed = BAD;
    }
#endif

    tiger_dbg (res, 0, nblocks, ncount);


    while (1) {

      count = sh_unix_fileread (fd, buffer, PRIV_MAX);

      blk      = (count / 64); /* number of 64-byte words */
      /* nblocks += blk; number of 64-byte words */
      ncount = count - ((count/64)*64);

      bptr = buffer;
      for (i = 0; i < blk; ++i)
	{
	  bptr = &buffer[64 * i];

	  tiger_t((sh_word32*) bptr, 64, res);
	  ++nblocks;
	  /*
	  tiger((sh_word32*) buffer, 
		( count < PRIV_MAX ? count : PRIV_MAX), res);
	  */
	  tiger_dbg (res, 3, nblocks, ncount);
	}

      if (blk < 8) 
	break;
    }

    /* copy incomplete block
     */
    j = 0; for (i = 0; i < 64; ++i) 
      bbuf[i] = '\0';
    for (i = (count/64) * 64; i < count; ++i)
      bbuf[j++] = buffer[i];

    /*
    for (i = (count/64)*64; i < PRIV_MAX; ++i) 
      buffer[i] = '\0';
    if ((count/64)*64 > 0)
      tiger_t((sh_word32*) buffer, (count/64)*64, res);
    */

    tiger_dbg (res, 5, nblocks, ncount);

    msb = 0;
    t = nblocks;
    if( (lsb = t << 6) < t )    
      msb++;
    msb += t >> 26;
    t = lsb;
    if( (lsb = t + ncount) < t ) 
      msb++;
    t = lsb;
    if( (lsb = t << 3) < t )    
      msb++;
    msb += t >> 29;

    if( j < 56 ) 
      { 
        bbuf[j++] = 0x01; ++ncount;
        while( j < 56 )
	  { bbuf[j++] = 0; ++ncount; } 
      }
    else 
      { 
        bbuf[j++] = 0x01;
        while( j < 64 )
	  bbuf[j++] = 0;
	tiger_t((sh_word32*) bbuf, 64, res);
	++nblocks; ncount = 0;
        memset(bbuf, 0, 56 ); 
      }

    tiger_dbg (res, 6, nblocks, ncount);

    bbuf[56] = lsb      ;
    bbuf[57] = lsb >>  8;
    bbuf[58] = lsb >> 16;
    bbuf[59] = lsb >> 24;
    bbuf[60] = msb      ;
    bbuf[61] = msb >>  8;
    bbuf[62] = msb >> 16;
    bbuf[63] = msb >> 24;

    tiger_t((sh_word32*) bbuf, 64, res);

    tiger_dbg (res, 7, nblocks, ncount);

    memset (bbuf,   0, 64);
    memset (buffer, 0, PRIV_MAX);

    MUNLOCK( buffer, (PRIV_MAX+2) * sizeof(sh_byte));

    sh_unix_fileclose (fd);

    sprintf(out,                                     /* known to fit  */
            _("%08lX%08lX %08lX%08lX %08lX%08lX"),
            (sh_word32)(res[1]), 
            (sh_word32)(res[0]), 
            (sh_word32)(res[3]), 
            (sh_word32)(res[2]), 
            (sh_word32)(res[5]), 
            (sh_word32)(res[4]) );
    out[50] = '\0';
    return (out);
 
  }


  if (what == TIGER_DATA) 
    {

      tiger((sh_word32*) filename,  (sh_word32) Length, res); 

      sprintf(out, _("%08lX%08lX %08lX%08lX %08lX%08lX"), 
	      (sh_word32)(res[1]), 
	      (sh_word32)(res[0]), 
	      (sh_word32)(res[3]), 
	      (sh_word32)(res[2]), 
	      (sh_word32)(res[5]), 
	      (sh_word32)(res[4]) );
      return (out);
    }

  return _("0000000000000000 000000000000000 000000000000000");

}

