/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */



#ifndef SH_UNIX_H
#define SH_UNIX_H

#include <limits.h>
#include <unistd.h>
#include "samhain.h"
#include "sh_error.h"


typedef enum {
  SH_ISLOG,
  SH_ISFILE,
  SH_ISDIR,
  SH_ISDATA
} ShOpenType;

typedef enum {
  SH_DATA_RAW,
  SH_DATA_LINE,
} ShDataType;

typedef enum {
  SH_FILE_REGULAR,
  SH_FILE_SYMLINK,
  SH_FILE_DIRECTORY,
  SH_FILE_CDEV,
  SH_FILE_BDEV,
  SH_FILE_FIFO,
  SH_FILE_SOCKET,
  SH_FILE_UNKNOWN
} ShFileType;


typedef struct file_struct {
  char             fullpath[PATH_MAX+2];
  ShFileType       type;
  unsigned long    dev;
  unsigned long    ino;
  unsigned int     mode;
  unsigned long    hardlinks;
  char             c_mode[11];
  long             owner;
  char             c_owner[OWNER_MAX+2];
  long             group;
  char             c_group[GROUP_MAX+2];
  unsigned long    rdev;
  unsigned long    size;
  unsigned long    blksize;
  unsigned long    blocks;
  unsigned long    atime;
  unsigned long    mtime;
  unsigned long    ctime;

  char             linkpath[PATH_MAX+2];
  unsigned int     linkmode;
  char             link_c_mode[11];
  int              linkisok;
} file_type;

unsigned long sh_unix_longtime ();
void sh_unix_closeall (int fd);

/* add a trusted user to the list 
 */
int tf_add_trusted_user(char *);

/* check a file 
 */
int tf_trust_check (char * file);

/* set the timeserver address
 */
int sh_unix_settimeserver (char * address);

/* lock the key
 */
void sh_unix_mlock();

/* deamon mode 
 */
int sh_unix_setdeamon(char * dummy);

/* test whether file exists
 */
int sh_unix_file_exists(char * file_name);

/* local host
 */
void sh_unix_localhost();

/* check whether /proc exists and is a proc filesystem
 */ 
int sh_unix_test_proc();

/* check whether a directory is secure 
 * (no symlink in path, not world-writeable)
 */
/* int sh_unix_is_secure_dir (ShErrLevel level, char * tmp); */

/* obtain file info
 */
int sh_unix_getinfo (int level,
		     char * filename, file_type * theFile);

/* unlock the default errlog file
 */
void sh_unix_unlock ();

/* lock a file
 */
void sh_unix_lock (char * s, char * lock);

/* privilege handling
 */
int sh_unix_isnotsuid(); /* return (ruid == euid) */
void sh_save_uids();
int sh_unset_suid ();
int sh_set_suid ();
int sh_unix_droppriv();

/* read file, return length read
 */
int sh_unix_getline (int fd, char * line, int sizeofline);

/* open() wrapper function
 */
int sh_unix_fileread (int fd, void *buf, size_t count);

/* call with goDaemon == 1 to make daemon process
 */
int  sh_unix_init(int goDaemon);

/* for local time use thetime = 0 
 */
char * sh_unix_time (unsigned long thetime);

/* effective user info
 */
int  sh_unix_getUser ();

/* open file with error checking, return file descriptor, (-1) on error
 */
int  sh_unix_fileopen (char * filename, char mode, 
		       ShOpenType  IsType, int level, int withpriv);

/* write to file with error checking
 */
int sh_unix_filewrite (int fd, void * msg, long bytecount, 
		       ShOpenType IsType, ShDataType IsData);

/* close file
 */
void sh_unix_fileclose (int fd);

/* check whether a directory
 */
int sh_unix_isdir (char * dirName, int level);

#ifdef SH_STEALTH
int  sh_unix_getline_stealth  (int fd, char * str, int len);
void sh_unix_xor_code (char * str, int len);
#endif

#endif
