#! /bin/sh


# The following two are the ANSI sequences for start and end embolden
case $TERM in
vt*|ansi*|con*|xterm*|linux*|screen*)
  S=[1m
  E=[m
  ;;
*)
  S=
  E=
  ;;
esac

PW_DIR=`pwd`

TRUST="--with-trusted=0,2"

echo; echo "${S}__ STARTING TEST SUITE __${E}"; echo;
echo Working directory: $PW_DIR

MAKE=`which gmake | sed -e "s%\([a-z:]\) .*%\1%g"` 
if test x$MAKE = x; then
	MAKE="make -s"
else
	if test "x$MAKE" = "xwhich:"; then
		MAKE="make -s"
	else
		MAKE="gmake -s"
	fi
fi
echo MAKE is $MAKE
echo

testgo ()
{
	#/bin/rm -f ./.samhain_*

	#echo; echo "${S}Initialize${E}: ./samhain -t init"; echo;
	#/bin/rm -f .samhain_file
	#./samhain -t init

	echo; echo "${S}Start Server${E}: ./yule -S -p none &"; echo;
	./yule -S -p none &
	PROC_Y=$!
	sleep 5

	echo; echo "${S}Start Agent${E}: ./samhain -l none -p none -e warn -t check &"; 
	echo;
	./samhain -l none -p none -e warn -t check &
	sleep 60
	PROC_S=$!
	
	kill ${PROC_S}
	sleep 5
	kill ${PROC_Y}

	./samhain -L ./.samhain_log
}

teststealth ()
{
	#
	# compilation
	#
	echo "${S}Building client/server application -- stealth mode ($1)${E}";
	echo;
	#
	if test -r "Makefile"; then
		$MAKE distclean
	fi
	#
	./configure --quiet $TRUST  --enable-srp --with-stealth=$1 --enable-login-watch --enable-network --prefix=$PW_DIR --with-dataroot-prefix=$PW_DIR --with-config-file=$PW_DIR/samhainrc.stealth 
	#
	if test x$? = x0; then
		echo "configure completed, exit status 0"; echo;
		$MAKE  > /dev/null
		out="$?"
		echo "$MAKE completed, exit status $out"; echo;
		if ! test x$out = x0 ; then
		    echo "ERROR: make failed";
		    exit
		fi
	else
		echo "ERROR: configure failed"; echo;
	fi

	/bin/cp samhainrc.test.stealth samhainrc.stealth
	./samhain_stealth -s samhainrc.stealth samhainrc.test 
	# run
	echo; echo "${S}Initialize${E}: ./samhain -t init"; echo;
	/bin/rm -f .samhain_file
	./samhain -t init

	echo; echo "${S}Check${E}: ./samhain -t check &"; echo;
	./samhain -t check &
	sleep 60
	kill $!
	sleep 5

	if test -r "./.samhain_lock"; then
	    echo "${S}removing ./.samhain_lock${E}"
	    /bin/rm -f ./.samhain_lock
	fi

	# cat ./.samhain_log
	/bin/rm -f ./.samhain_log

	echo; echo "${S}Start Server${E}: ./yule -S -p none &"; echo;
	./yule -S -p none &
	PROC_Y=$!
	sleep 5

	echo; echo "${S}Start Agent${E}: ./samhain -l none -p none -e warn -t check &"; 
	echo;
	./samhain -l none -p none -e warn  -t check &
	sleep 60
	PROC_S=$!
	
	kill ${PROC_S}
	sleep 5
	kill ${PROC_Y}

	./samhain -L ./.samhain_log
}

testbuild ()
{
	#
	# compilation
	#
	echo "${S}Building client/server application${E}"; echo;
	#
	if test -r "Makefile"; then
		$MAKE distclean
	fi
	#
	./configure --quiet $TRUST --enable-srp --enable-login-watch --enable-network --prefix=$PW_DIR --with-dataroot-prefix=$PW_DIR --with-config-file=$PW_DIR/samhainrc.test   
	#
	if test x$? = x0; then
		echo "configure completed, exit status 0"; echo;
		$MAKE  samhain > /dev/null
		out="$?"
		echo "$MAKE completed, exit status $out"; echo;
		if ! test x$out = x0 ; then
		    echo "ERROR: make failed";
		    exit
		fi
	else
		echo "ERROR: configure failed"; echo;
	fi
}


testrun ()
{
	#
	# compilation
	#
	echo "${S}Building client/server application${E}"; echo;
	#
	if test -r "Makefile"; then
		$MAKE distclean
	fi
	#
	./configure --quiet $TRUST --enable-srp --enable-debug --enable-network --prefix=$PW_DIR --with-dataroot-prefix=$PW_DIR --with-config-file=$PW_DIR/samhainrc.test   
	#
	if test x$? = x0; then
		echo "configure completed, exit status 0"; echo;
		$MAKE  > /dev/null
		out="$?"
		echo "$MAKE completed, exit status $out"; echo;
		if ! test x$out = x0 ; then
		    echo "ERROR: make failed";
		    exit 8
		fi
	else
		echo "ERROR: configure failed"; echo;
	fi

	# run
	echo; echo "${S}Initialize:${E} ./samhain -t init -d 1"; echo;
	/bin/rm -f .samhain_*
	./samhain -t init -d 1

	echo; echo "${S}Check:${E} ./samhain -t check -d 1 &"; echo;
	./samhain -t check -d 1 &
	PROC_S=$!
	sleep 60
	kill ${PROC_S}
	sleep 5

	if test -r "./.samhain_lock"; then
	    echo "${S}removing ./.samhain_lock${E}"
	    /bin/rm -f ./.samhain_lock
	fi

	cat ./.samhain_log
	/bin/rm -f ./.samhain_log

	echo; echo "${S}Start Server:${E} ./yule -S -p none  &"; echo;
	./yule -S -p none &
	PROC_Y=$!
	sleep 5

	echo; 
	echo "${S}Start Agent: ${E} ./samhain  -l none -p none -e warn -t check &";
	echo;
	./samhain -l none -p none -e warn  -t check &
	sleep 60
	PROC_S=$!
	
	kill ${PROC_S}
	sleep 5
	kill ${PROC_Y}

	cat ./.samhain_log

	echo
	echo "${S}Testing the hash algorith${E}: ./samhain -H Tiger"; echo;
	echo "EXPECT |9F00F599072300DD 276ABB38C8EB6DEC 37790C116F9D2BDF|"
	echo "run command ..."
	./samhain -H Tiger
}


testcompile ()
{
	#
	# test standalone compilation
	#
	echo "${S}Building standalone agent${E}"; echo;
	#
	if test -r "Makefile"; then
		$MAKE distclean
	fi
	#
	./configure --quiet  --enable-srp --prefix=$PW_DIR --with-dataroot-prefix=$PW_DIR --with-config-file=$PW_DIR/samhainrc.test   
	#
	if test x$? = x0; then
		echo "configure completed, exit status 0"; echo;
		$MAKE  > /dev/null
		echo "$MAKE completed, exit status $?"; echo;
	else
		echo "ERROR: configure failed"; echo;
	fi

	#
	# test standalone compilation w/ debug
	#
	echo "${S}Building standalone agent w/debug${E}"; echo;
	#
	if test -r "Makefile"; then
		$MAKE distclean
	fi
	#
	./configure --quiet --enable-debug  --enable-srp --prefix=$PW_DIR --with-dataroot-prefix=$PW_DIR --with-config-file=$PW_DIR/samhainrc.test   
	#
	if test x$? = x0; then
		echo "configure completed, exit status 0"; echo;
		$MAKE  > /dev/null
		echo "$MAKE completed, exit status $?"; echo;
	else
		echo "ERROR: configure failed"; echo;
	fi

	#
	# test standalone compilation w/stealth
	#
	echo "${S}Building standalone agent w/stealth${E}"; echo;
	#
	if test -r "Makefile"; then
		$MAKE distclean
	fi
	#
	./configure --quiet --with-stealth=128 --prefix=$PW_DIR --with-dataroot-prefix=$PW_DIR --with-config-file=$PW_DIR/samhainrc.test   
	#
	if test x$? = x0; then
		echo "configure completed, exit status 0"; echo;
		$MAKE  > /dev/null
		echo "$MAKE completed, exit status $?"; echo;
	else
		echo "ERROR: configure failed"; echo;
	fi

	#
	# test standalone compilation w/logwatch
	#
	echo "${S}Building standalone agent w/login-watch${E}"; echo;
	#
	if test -r "Makefile"; then
		$MAKE distclean
	fi
	#
	./configure --quiet --enable-login-watch  --enable-srp --prefix=$PW_DIR --with-dataroot-prefix=$PW_DIR --with-config-file=$PW_DIR/samhainrc.test   
	#
	if test x$? = x0; then
		echo "configure completed, exit status 0"; echo;
		$MAKE  > /dev/null
		echo "$MAKE completed, exit status $?"; echo;
	else
		echo "ERROR: configure failed"; echo;
	fi

	#
	# test standalone compilation
	#
	echo "${S}Building standalone agent w/stealth and debug${E}"; echo;
	#
	if test -r "Makefile"; then
		$MAKE distclean
	fi
	#
	./configure --quiet --with-stealth=128  --enable-srp --enable-debug --prefix=$PW_DIR --with-dataroot-prefix=$PW_DIR --with-config-file=$PW_DIR/samhainrc.test   
	#
	if test x$? = x0; then
		echo "configure completed, exit status 0"; echo;
		$MAKE  > /dev/null
		echo "$MAKE completed, exit status $?"; echo;
	else
		echo "ERROR: configure failed"; echo;
	fi

	echo; echo "${S}__ TEST CLIENT/SERVER __${E}"; echo;

	#
	# test client/server compilation
	#
	echo "${S}Building client/server application${E}"; echo;
	#
	if test -r "Makefile"; then
		$MAKE distclean
	fi
	#
	./configure --quiet --enable-network  --enable-srp --prefix=$PW_DIR --with-dataroot-prefix=$PW_DIR --with-config-file=$PW_DIR/samhainrc.test   
	#
	if test x$? = x0; then
		echo "configure completed, exit status 0"; echo;
		$MAKE  > /dev/null
		echo "$MAKE completed, exit status $?"; echo;
	else
		echo "ERROR: configure failed"; echo;
	fi

	#
	# test client/server compilation w/ debug
	#
	echo "${S}Building client/server application w/debug${E}"; echo;
	#
	if test -r "Makefile"; then
		$MAKE distclean
	fi
	#
	./configure --quiet --enable-network --enable-debug --prefix=$PW_DIR --with-dataroot-prefix=$PW_DIR --with-config-file=$PW_DIR/samhainrc.test   
	#
	if test x$? = x0; then
		echo "configure completed, exit status 0"; echo;
		$MAKE  > /dev/null
		echo "$MAKE completed, exit status $?"; echo;
	else
		echo "ERROR: configure failed"; echo;
	fi

	#
	# test client/server compilation w/stealth
	#
	echo "${S}Building client/server application w/stealth${E}"; echo;
	#
	if test -r "Makefile"; then
		$MAKE distclean
	fi
	#
	./configure --quiet --enable-network  --enable-srp --with-stealth=128 --prefix=$PW_DIR --with-dataroot-prefix=$PW_DIR --with-config-file=$PW_DIR/samhainrc.test   
	#
	if test x$? = x0; then
		echo "configure completed, exit status 0"; echo;
		$MAKE  > /dev/null
		echo "$MAKE completed, exit status $?"; echo;
	else
		echo "ERROR: configure failed"; echo;
	fi

	#
	# test  client/server compilation w/logwatch
	#
	echo "${S}Building client/server application w/login-watch${E}"; echo;
	#
	if test -r "Makefile"; then
		$MAKE distclean
	fi
	#
	./configure --quiet --enable-network  --enable-srp --enable-login-watch --prefix=$PW_DIR --with-dataroot-prefix=$PW_DIR --with-config-file=$PW_DIR/samhainrc.test   
	#
	if test x$? = x0; then
		echo "configure completed, exit status 0"; echo;
		$MAKE  > /dev/null
		echo "$MAKE completed, exit status $?"; echo;
	else
		echo "ERROR: configure failed"; echo;
	fi

	#
	# test client/server compilation
	#
	echo "${S}Building client/server application w/stealth and debug${E}";
	echo;
	#
	if test -r "Makefile"; then
		$MAKE distclean
	fi
	#
	./configure --quiet --enable-network  --enable-srp --with-stealth=128 --enable-debug --prefix=$PW_DIR --with-dataroot-prefix=$PW_DIR --with-config-file=$PW_DIR/samhainrc.test   
	#
	if test x$? = x0; then
		echo "configure completed, exit status 0"; echo;
		$MAKE  > /dev/null
		echo "$MAKE completed, exit status $?"; echo;
	else
		echo "ERROR: configure failed"; echo;
	fi
}


if test "x$1" = "x"; then
	echo "Usage: testconf.sh compile|run|stealth|build "
	echo "  - compile - Test a variety of compile options: "
	echo "              vanilla, debug, stealth, all three"
	echo "              in standalone/network mode."
	echo "  - run     - Compile, initialize database, check,"
	echo "              do local client/server communication."
	echo "  - stealth - Same as 'run', but in stealth mode."
	echo "  - build   - Just build it."
	echo "  - go      - Just run it."
	echo
	echo "  CHECK samhainrc.test BEFORE USING run|stealth"
fi

if ! test -r samhainrc.test ; then
	cp samhainrc samhainrc.test
fi

if test x$2 = x; then
	vv=163
fi

if test x$1 = xcompile; then
	testcompile
fi
if test x$1 = xrun; then
	testrun
fi
if test x$1 = xstealth; then
	teststealth $vv
fi
if test x$1 = xbuild; then
	testbuild
fi
if test x$1 = xgo; then
	testgo
fi

echo; echo "${S}__ END TEST SUITE __${E}"; echo;

exit

